"""
PPM Utils
"""

__title__ = "PPM Utils"
__version__ = "1.0.0-beta.6"
__author__ = "Bryan Larson"
__license__ = "BSD 2-Clause"
__copyright__ = "Copyright 2011-2020 Harvard Medical School Department of Biomedical Informatics"

# Version synonym
VERSION = __version__

# Header encoding (see RFC5987)
HTTP_HEADER_ENCODING = "iso-8859-1"

# Default datetime input and output formats
ISO_8601 = "iso-8601"

# Set the Django app
default_app_config = "ppmutils.apps.PPMUtilsConfig"
