<!DOCTYPE html>
<html>
<head>
    <meta name="viewport" content="width=device-width">
    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
    <title>{%block title%}{%endblock%}</title>
    <style type="text/css">
    /* -------------------------------------
        INLINED WITH https://putsmail.com/inliner
    ------------------------------------- */
    /* -------------------------------------
        RESPONSIVE AND MOBILE FRIENDLY STYLES
    ------------------------------------- */
    @media only screen and (max-width: 620px) {
      table[class=body] h1 {
        font-size: 28px !important;
        margin-bottom: 10px !important; }
      table[class=body] p,
      table[class=body] ul,
      table[class=body] ol,
      table[class=body] td,
      table[class=body] span,
      table[class=body] a {
        font-size: 16px !important; }
      table[class=body] .wrapper,
      table[class=body] .article {
        padding: 10px !important; }
      table[class=body] .content {
        padding: 0 !important; }
      table[class=body] .container {
        padding: 0 !important;
        width: 100% !important; }
      table[class=body] .main {
        border-left-width: 0 !important;
        border-radius: 0 !important;
        border-right-width: 0 !important; }
      table[class=body] .btn table {
        width: 100% !important; }
      table[class=body] .btn a {
        width: 100% !important; }
      table[class=body] .img-responsive {
        height: auto !important;
        max-width: 100% !important;
        width: auto !important; }}
    /* -------------------------------------
        PRESERVE THESE STYLES IN THE HEAD
    ------------------------------------- */
    @media all {
      .ExternalClass {
        width: 100%; }
      .ExternalClass,
      .ExternalClass p,
      .ExternalClass span,
      .ExternalClass font,
      .ExternalClass td,
      .ExternalClass div {
        line-height: 100%; }
      .apple-link a {
        color: inherit !important;
        font-family: inherit !important;
        font-size: inherit !important;
        font-weight: inherit !important;
        line-height: inherit !important;
        text-decoration: none !important; }
      .btn-primary table td:hover {
        background-color: #34495e !important; }
      .btn-primary a:hover {
        background-color: #34495e !important;
        border-color: #34495e !important; } }

    </style>
</head>
<body class=""
      style="background-color:#f6f6f6;font-family:sans-serif;-webkit-font-smoothing:antialiased;font-size:14px;line-height:1.4;margin:0;padding:0;-ms-text-size-adjust:100%;-webkit-text-size-adjust:100%;">
<table border="0" cellpadding="0" cellspacing="0" class="body"
       style="border-collapse:separate;mso-table-lspace:0pt;mso-table-rspace:0pt;background-color:#f6f6f6;width:100%;">
    <tr>
        <td style="font-family:sans-serif;font-size:14px;vertical-align:top;">&nbsp;</td>
        <td class="container"
            style="font-family:sans-serif;font-size:14px;vertical-align:top;display:block;max-width:580px;padding:10px;width:580px;Margin:0 auto !important;">
            <div class="content"
                 style="box-sizing:border-box;display:block;Margin:0 auto;max-width:580px;padding:10px;">
                <!-- START CENTERED WHITE CONTAINER -->
                <span class="preheader"
                      style="color:transparent;display:none;height:0;max-height:0;max-width:0;opacity:0;overflow:hidden;mso-hide:all;visibility:hidden;width:0;">People-Powered Medicine - {{ ppm_study_title }} - Notification</span>
                <table class="main"
                       style="border-collapse:separate;mso-table-lspace:0pt;mso-table-rspace:0pt;background:#fff;border-radius:3px;width:100%;">
                    <!-- START MAIN CONTENT AREA -->
                    <tr>
                        <td class="wrapper"
                            style="font-family:sans-serif;font-size:14px;vertical-align:top;box-sizing:border-box;padding:20px;">
                            <table border="0" cellpadding="0" cellspacing="0"
                                   style="border-collapse:separate;mso-table-lspace:0pt;mso-table-rspace:0pt;width:100%;">
                                <tr>
                                    <td style="font-family:sans-serif;font-size:14px;vertical-align:top;box-sizing:border-box;padding:20px;border-bottom-color:#eeeeee;border-bottom-width:1px;border-bottom-style:solid;">
                                        <img src="https://peoplepoweredmedicine.org/img/ppm_logo-120x35.png"/>
                                    </td>
                                </tr>
                                <tr>
                                    <td style="font-family:sans-serif;font-size:14px;vertical-align:top;padding:10px 20px;">
                                        <p>{% block greeting %}{% if first_name %}{{ first_name }}{% else %}Hello{% endif %},{% endblock %}</p>
                                    </td>
                                </tr>
                                <tr>
                                    <td style="font-family:sans-serif;font-size:14px;vertical-align:top;padding:0 20px;">
                                        {% block content %}{% endblock %}
                                    </td>
                                </tr>
                                <tr>
                                <td style="font-family:sans-serif;font-size:14px;vertical-align:top;box-sizing:border-box;padding:20px;">
                                    {% block signature %}
                                    {% if ppm_signature %}
                                    <table border="0" cellpadding="0" cellspacing="0"
                                           style="border-collapse:separate;mso-table-lspace:0pt;mso-table-rspace:0pt;width:100%;">
                                        <tr>
                                            <td class="content-block" style="font-family:sans-serif;vertical-align:top;color:#000000;font-size:16px;padding-bottom: 5px;">
                                                {{ ppm_signature.name }}
                                            </td>
                                        </tr>
                                        <tr>
                                            <td class="content-block" style="font-family:sans-serif;vertical-align:top;color:#555555;font-size:14px;">
                                                {{ ppm_signature.title }}
                                            </td>
                                        </tr>
                                        <tr>
                                            <td class="content-block" style="font-family:sans-serif;vertical-align:top;color:#999999;font-size:12px;">
                                                Harvard Medical School
                                            </td>
                                        </tr>
                                        <tr>
                                            <td class="content-block" style="font-family:sans-serif;vertical-align:top;color:#999999;font-size:12px;">
                                                10 Shattuck Street
                                            </td>
                                        </tr>
                                        <tr>
                                            <td class="content-block" style="font-family:sans-serif;vertical-align:top;color:#999999;font-size:12px;">
                                                Boston, MA 02115
                                            </td>
                                        </tr>
                                        <tr>
                                            <td class="content-block" style="font-family:sans-serif;vertical-align:top;color:#555555;font-size:12px;">
                                                <a href="mailto:{{ ppm_signature.email }}">{{ ppm_signature.email }}</a>
                                            </td>
                                        </tr>
                                        <tr>
                                            <td class="content-block" style="font-family:sans-serif;vertical-align:top;color:#555555;font-size:12px;">
                                                Ph: {{ ppm_signature.phone }}
                                            </td>
                                        </tr>
                                    </table>
                                    {% endif %}
                                    {% endblock %}
                                </td>
                            </tr>
                            </table>
                        </td>
                    </tr>
                    <!-- END MAIN CONTENT AREA -->
                </table>
                <!-- START FOOTER -->
                <div class="footer" style="clear:both;padding-top:10px;text-align:center;width:100%;">
                    <table border="0" cellpadding="0" cellspacing="0"
                           style="border-collapse:separate;mso-table-lspace:0pt;mso-table-rspace:0pt;width:100%;">
                        <tr>
                            <td class="content-block powered-by"
                                style="font-family:sans-serif;font-size:14px;vertical-align:top;color:#999999;font-size:12px;text-align:center;">Sent by
                                <a href="https://peoplepoweredmedicine.org"
                                   style="color:#3498db;text-decoration:underline;color:#3498db;font-size:12px;text-align:center;text-decoration:none;">People-Powered
                                    Medicine</a>.
                            </td>
                        </tr>
                        <tr>
                            <td class="content-block"
                                style="font-family:sans-serif;font-size:14px;vertical-align:top;color:#999999;font-size:12px;text-align:center;">
                                <span class="apple-link" style="color:#999999;font-size:12px;text-align:center;">Harvard Medical School, 10 Shattuck St, Boston, MA 02115</span>
                            </td>
                        </tr>
                    </table>
                </div>
                <!-- END FOOTER -->
                <!-- END CENTERED WHITE CONTAINER -->
            </div>
        </td>
        <td style="font-family:sans-serif;font-size:14px;vertical-align:top;">&nbsp;</td>
    </tr>
</table>
</body>
</html>