{% extends "ppmutils/_base.html" %}

{% block title %}{{ ppm_subject }}{% endblock %}

{% block content %}
<p style="font-family:sans-serif;font-size:14px;font-weight:normal;margin:0;Margin-bottom:15px;">Thank you for your patience as we review your registration for the People-Powered Medicine {{ ppm_study_title }} study.  We wanted to inform you that we are still examining the registration information that you provided to determine eligibility and are consulting with additional oncology experts.  We may reach out to you as we move forward to clarify any potential questions we may have.  Please don’t hesitate to contact us if you have any questions or concerns.</p>
<p style="font-family:sans-serif;font-size:14px;font-weight:normal;margin:0;Margin-bottom:15px;">Thank you for your continued interest in PPM {{ ppm_study_title }}!</p>
<br />
<table border="0" cellpadding="0" cellspacing="0" class="btn btn-primary" style="border-collapse:separate;mso-table-lspace:0pt;mso-table-rspace:0pt;box-sizing:border-box;width:100%;">
    <tbody>
        <tr>
            <td align="left" style="font-family:sans-serif;font-size:14px;vertical-align:top;padding-bottom:15px;">
                <table border="0" cellpadding="0" cellspacing="0" style="border-collapse:separate;mso-table-lspace:0pt;mso-table-rspace:0pt;width:100%;width:auto;">
                    <tbody>
                        <tr>
                            <td style="font-family:sans-serif;font-size:14px;vertical-align:top;background-color:#ffffff;border-radius:5px;text-align:center;background-color:#3498db;">
                            <a href="{{ dashboard_url }}" target="_blank" style="text-decoration:underline;background-color:#ffffff;border:solid 1px #3498db;border-radius:5px;box-sizing:border-box;color:#3498db;cursor:pointer;display:inline-block;font-size:14px;font-weight:bold;margin:0;padding:12px 25px;text-decoration:none;text-transform:capitalize;background-color:#3498db;border-color:#3498db;color:#ffffff;">
                                Visit your dashboard</a>
                            </td>
                        </tr>
                    </tbody>
                </table>
            </td>
        </tr>
    </tbody>
</table>
<br />
{% endblock %}
