{% extends "ppmutils/_base.html" %}

{% block title %}{{ ppm_subject }}{% endblock %}

{% block content %}
<p style="font-family:sans-serif;font-size:14px;font-weight:normal;margin:0;Margin-bottom:15px;">We have reviewed your registration information for the People-Powered Medicine {{ ppm_study_title }} study and are pleased to offer you enrollment as a participant.  In the next several weeks, you will receive your materials (for example, a Fitbit if you opted to wear one) for participation in the study as well as instructions for use.  We will also reach out to you for any required signatures as we begin to collect your medical records and access previous pathology samples for DNA sequencing.  Please don’t hesitate to contact us if you have any questions or concerns about any part of the study.</p>
<p style="font-family:sans-serif;font-size:14px;font-weight:normal;margin:0;Margin-bottom:15px;">If you no longer wish to be enrolled in this study, please contact us as soon as possible and we will remove your information from the system.</p>
<p style="font-family:sans-serif;font-size:14px;font-weight:normal;margin:0;Margin-bottom:15px;">Thank you for sharing your remarkable experience with us, and we look forward to having you join us for this valuable research!</p>
<br />
<table border="0" cellpadding="0" cellspacing="0" class="btn btn-primary" style="border-collapse:separate;mso-table-lspace:0pt;mso-table-rspace:0pt;box-sizing:border-box;width:100%;">
    <tbody>
        <tr>
            <td align="left" style="font-family:sans-serif;font-size:14px;vertical-align:top;padding-bottom:15px;">
                <table border="0" cellpadding="0" cellspacing="0" style="border-collapse:separate;mso-table-lspace:0pt;mso-table-rspace:0pt;width:100%;width:auto;">
                    <tbody>
                        <tr>
                            <td style="font-family:sans-serif;font-size:14px;vertical-align:top;background-color:#ffffff;border-radius:5px;text-align:center;background-color:#3498db;">
                            <a href="{{ dashboard_url }}" target="_blank" style="text-decoration:underline;background-color:#ffffff;border:solid 1px #3498db;border-radius:5px;box-sizing:border-box;color:#3498db;cursor:pointer;display:inline-block;font-size:14px;font-weight:bold;margin:0;padding:12px 25px;text-decoration:none;text-transform:capitalize;background-color:#3498db;border-color:#3498db;color:#ffffff;">
                                Visit your dashboard</a>
                            </td>
                        </tr>
                    </tbody>
                </table>
            </td>
        </tr>
    </tbody>
</table>
<br />
{% endblock %}