{% extends "ppmutils/_base.html" %}

{% block title %}{{ ppm_subject }}{% endblock %}

{% block greeting %}{% endblock %}

{% block content %}

<p style="font-family:sans-serif;font-size:16px;font-weight:700;margin:0;Margin-bottom:15px;">From: <span style="font-family:sans-serif;font-size:14px;font-weight:normal;margin:0;Margin-bottom:15px;">{{ from_name }}</span></p>
<p style="font-family:sans-serif;font-size:16px;font-weight:700;margin:0;Margin-bottom:15px;">Email: <span style="font-family:sans-serif;font-size:14px;font-weight:normal;margin:0;Margin-bottom:15px;">{{ from_email}}</p>
<p style="font-family:sans-serif;font-size:16px;font-weight:700;margin:0;Margin-bottom:15px;">Study: <span style="font-family:sans-serif;font-size:14px;font-weight:normal;margin:0;Margin-bottom:15px;">{{ ppm_study_title }}</p>
<p style="font-family:sans-serif;font-size:16px;font-weight:700;margin:0;Margin-bottom:15px;">Comment:</p>
<p style="font-family:sans-serif;font-size:14px;font-weight:normal;margin:0;Margin-bottom:15px;">{{ message }}</p>

{% endblock %}

{% block signature %}{% endblock %}